# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from datetime import timedelta
from typing import Union

import numpy as np
import pandas as pd
import structlog

from openstef.preprocessing.preprocessing import replace_repeated_values_with_nan
from openstef.model.regressors.regressor import OpenstfRegressor


def validate(
    pj_id: Union[int, str],
    data: pd.DataFrame,
    flatliner_threshold: Union[int, None],
) -> pd.DataFrame:
    """Validate prediction job and timeseries data.

    Steps:
    1. Replace repeated values for longer than flatliner_threshold with NaN
    # TODO: The function description suggests it
    'validates' the PJ and Data, but is appears to 'just' replace repeated observations with NaN.

    Args:
        pj_id: ind/str, used to identify log statements
        data: pd.DataFrame where the first column should be the target. index=datetimeIndex
        flatliner_threshold: int of max repetitions considered a flatline.
            if None, the validation is effectively skipped

    Returns:
        Dataframe where repeated values are set to None

    """
    logger = structlog.get_logger(__name__)

    # Check if DataFrame has datetime index
    if not isinstance(data.index, pd.DatetimeIndex):
        raise ValueError("Input dataframe does not have a datetime index.")

    if flatliner_threshold is None:
        logger.info("Skipping validation of input data", pj_id=pj_id)
        return data

    # Drop 'false' measurements. e.g. where load appears to be constant.
    data = replace_repeated_values_with_nan(
        data, max_length=flatliner_threshold, column_name=data.columns[0]
    )
    num_repeated_values = len(data) - len(data.iloc[:, 0].dropna())
    if num_repeated_values > 0:
        frac_const_load_values = num_repeated_values / len(data.index)

        logger.info(
            f"Found {num_repeated_values} values of constant load (repeated values),"
            " converted to NaN value.",
            cleansing_step="repeated_values",
            pj_id=pj_id,
            num_values=num_repeated_values,
            frac_values=frac_const_load_values,
        )

    return data


def drop_target_na(data: pd.DataFrame) -> pd.DataFrame:
    logger = structlog.get_logger(__name__)
    len_original = len(data)
    # Remove where load is NA, NaN features are preserved
    data = data.loc[np.isnan(data.iloc[:, 0]) != True, :]  # noqa E712
    num_removed_values = len_original - len(data)
    if num_removed_values > 0:
        logger.info(
            f"Removed {num_removed_values} NaN values",
            num_removed_values=num_removed_values,
        )

    return data


def is_data_sufficient(
    data: pd.DataFrame,
    completeness_threshold: float,
    minimal_table_length: int,
    model: OpenstfRegressor = None,
) -> bool:
    """Check if enough data is left after validation and cleaning to continue with model training.

    Args:
        data: pd.DataFrame() with cleaned input data.
        model: model which contains all information regarding trained model
        completeness_threshold: float with threshold for completeness:
            1 for fully complete, 0 for anything could be missing.
        minimal_table_length: int with minimal table length (in rows)

    Returns:
        True if amount of data is sufficient, False otherwise.

    """
    if model is None:
        weights = None  # Remove horizon & load column
    else:
        weights = model.feature_importance_dataframe

    logger = structlog.get_logger(__name__)
    # Set output variable
    is_sufficient = True

    # Calculate completeness
    completeness = calc_completeness_features(
        data, weights, time_delayed=True, homogenise=False
    )
    table_length = data.shape[0]

    # Check if completeness is up to the standards
    if completeness < completeness_threshold:
        logger.warning(
            "Input data is not sufficient, completeness too low",
            completeness=completeness,
            completeness_threshold=completeness_threshold,
        )
        is_sufficient = False

    # Check if absolute amount of rows is sufficient
    if table_length < minimal_table_length:
        logger.warning(
            "Input data is not sufficient, table length too short",
            table_length=table_length,
            table_length_threshold=minimal_table_length,
        )
        is_sufficient = False

    return is_sufficient


def calc_completeness_features(
    df: pd.DataFrame,
    weights: pd.DataFrame,
    time_delayed: bool = False,
    homogenise: bool = True,
) -> float:
    """Calculate the (weighted) completeness of a dataframe.

    NOTE: NA values count as incomplete

    Args:
        df: Dataframe with a datetimeIndex index
        weights: Array-compatible with size equal to columns of df
            (excl. load&horizon), used to weight the completeness of each column
        time_delayed: Should there be a correction for T-x columns
        homogenise: Should the index be resampled to median time delta -
            only available for DatetimeIndex

    Returns:
        Fraction of completeness

    """
    df_copy = df.copy(
        deep=True
    )  # Make deep copy to maintain original dataframe in pipeline

    # Remove load and horizon from data_with_features dataframe to make sure columns are equal
    if "load" in df_copy:
        df_copy.drop("load", inplace=True, axis=1)
    if "horizon" in df_copy:
        df_copy.drop("horizon", inplace=True, axis=1)

    if weights is None:
        weights = np.array([1] * ((len(df_copy.columns))))

    length_weights = len(weights)
    length_features = len(df_copy.columns)

    # Returns the list
    if type(weights) != np.ndarray:
        list_features = weights.index.tolist()
        df_copy = df_copy[list_features]  # Reorder the df to match weights index (list)
        weights = weights.weight

    if length_weights != length_features:
        raise ValueError(
            "Input data is not sufficient, number of features used in model is unequal to amount of columns in data"
        )
    completeness_per_column_dataframe = calc_completeness_dataframe(
        df_copy, time_delayed, homogenise
    )

    # scale to weights and normalize
    completeness = (completeness_per_column_dataframe * weights).sum() / weights.sum()

    return completeness


def find_nonzero_flatliner(
    df: pd.DataFrame, threshold: int = None
) -> Union[pd.DataFrame, None]:
    """Function that detects a stationflatliner and returns a list of datetimes.

    Args:
        df: Example pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]).
            Load_corrections should be indicated by 'LC_'
        threshold: after how many timesteps should the function detect a flatliner.
            If None, the check is not executed

    Returns:
        Flatline moments or None

    TODO: a lot of the logic of this function can be improved using: mnts.label

    ```
    import scipy.ndimage.measurements as mnts
    mnts.label
    ```

    """
    if len(df) == 0:
        return

    # remove load corrections
    df = df.loc[:, ~df.columns.str.startswith("LC_")]

    # Remove moments when total load is 0
    df = df[(df.T != 0).any()]

    # Add columns with diff
    diff_columns = ["diff_" + x for x in df.columns]
    df[diff_columns] = df.diff().abs()
    # We are looking for when total station is a flatliner, so all
    # Select where all diff columns are zero
    flatliner_total = (df[diff_columns] == 0).mean(axis=1) == 1
    # Find all first and last occurences of 0
    first_zeros = flatliner_total[:-1][
        (flatliner_total.iloc[:-1] == 0).values * (flatliner_total.iloc[1:] != 0).values
    ]
    last_zeros = flatliner_total[1:][
        (flatliner_total.iloc[:-1] != 0).values * (flatliner_total.iloc[1:] == 0).values
    ]

    # If a zero-value is at start or end of df, remove from last_* list
    if len(flatliner_total) > 0:
        if flatliner_total.iloc[0]:
            last_zeros = last_zeros[1:]
        if flatliner_total.iloc[-1]:
            first_zeros = first_zeros[:-1]

    # Give as output from:to
    interval_df = pd.DataFrame(
        {
            "from_time": first_zeros.index,
            "to_time": last_zeros.index,
            "duration_h": last_zeros.index - first_zeros.index,
        },
        index=range(len(first_zeros)),
    )

    # Only keep periods which exceed threshold
    interval_df = interval_df[interval_df["duration_h"] >= timedelta(hours=threshold)]
    if len(interval_df) == 0:
        interval_df = None
    return interval_df


def find_zero_flatliner(
    df: pd.DataFrame,
    threshold: float,
    flatliner_window: timedelta,
    flatliner_load_threshold: float,
) -> pd.DataFrame or None:
    """Detect a zero value where the load is not compensated by the other trafo's of the station.

    If zero value is at start or end, ignore that block.

    Args:
        df: DataFrame such as pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]). Load_corrections should be indicated by 'LC_'
        threshold: after how many hours should the function detect a flatliner.
        flatliner_window: for how many hours before the zero-value should the mean load be calculated.
        flatliner_load_threshold: how big may the difference be between the total station load
            before and during the zero-value(s).

    Return:
        DataFrame of timestamps, or None if none

    TODO: a lot of the logic of this function can be improved using: mnts.label
    ```
    import scipy.ndimage.measurements as mnts
    mnts.label
    ```

    """
    result_df = pd.DataFrame()

    for col in df.columns:
        # Check for each trafo if it contains zero-values or NaN values
        if check_data_for_each_trafo(df, col):
            # Copy column in order to manipulate data
            df_col = df[col].copy()
            # Create list of zero_values to check length
            zero_values = df_col == 0
            last_zeros = zero_values[:-1][
                (zero_values.iloc[:-1] != 0).values * (zero_values.iloc[1:] == 0).values
            ]
            first_zeros = zero_values[1:][
                (zero_values.iloc[:-1] == 0).values * (zero_values.iloc[1:] != 0).values
            ]

            # If a zero-value is at start or end of df, remove from last_* list
            if len(zero_values) > 0:
                if zero_values.iloc[0]:
                    last_zeros = last_zeros[1:]
                if zero_values.iloc[-1]:
                    first_zeros = first_zeros[:-1]

            interval_df = pd.DataFrame(
                {
                    "from_time": first_zeros.index,
                    "to_time": last_zeros.index,
                    "duration_h": last_zeros.index - first_zeros.index,
                },
                index=range(len(first_zeros)),
            )

            # Keep dataframe with zero_value > threshold
            interval_df = interval_df[
                interval_df["duration_h"] >= timedelta(hours=threshold)
            ]

            # Check if there are missing values in the data
            num_nan = np.sum(pd.isna(df_col))
            if np.sum(num_nan) > 0:
                print(
                    "Found {a} missing values at trafo {b}, dropping NaN values".format(
                        a=num_nan, b=col
                    )
                )
                df_col = df_col.dropna()

            non_compensated_df = pd.DataFrame()
            print(
                "Found {b} suspicious moments were load is a zero at trafo {a}".format(
                    a=col, b=len(interval_df)
                )
            )
            # Forloop all moments of zero_values, check if delta of to_zero_trafo is compesated by other trafos
            for i, candidate in interval_df.iterrows():
                # Calculate mean load before and after zero-moment
                mean_load_before = (
                    df[candidate.from_time - flatliner_window : candidate.from_time]
                    .sum(axis=1)
                    .mean()
                )
                mean_full_timespan = (
                    df[candidate.from_time : candidate.from_time + flatliner_window]
                    .sum(axis=1)
                    .mean()
                )

                # Compare load and detect zero-value flatliner
                if (
                    np.abs(
                        (mean_load_before - mean_full_timespan)
                        / np.max([np.abs(mean_load_before), np.abs(mean_full_timespan)])
                    )
                    > flatliner_load_threshold
                ):
                    transposed_candidate = candidate.to_frame().T
                    non_compensated_df = pd.concat(
                        [non_compensated_df, transposed_candidate]
                    )
                    print(
                        "Found a non-compensated zero value:",
                        candidate.to_string(index=False),
                    )

            # after checking all candidates for a single trafo, add moments to result_df
            result_df = pd.concat([result_df, non_compensated_df])

    if len(result_df) == 0:
        result_df = None
    return result_df


def check_data_for_each_trafo(df: pd.DataFrame, col: pd.Series) -> bool:
    """Function that detects if each column contains zero-values at all, only zero-values and NaN values.

    Args:
        df: DataFrama such as pd.dataFrame(index=DatetimeIndex, columns = [load1, ..., loadN]).
            Load_corrections should be indicated by 'LC_'
        col: column of pd.dataFrame

    Returns:
        False if column contains above specified or True if not

    """
    if df is None:
        return False

    # Check for each column the data on the following: (Skipping if true)
    # Check if there a zero-values at all
    if (df[col] != 0).all(axis=0):
        print(f"No zero values found - at all at trafo {col}, skipping column")
        return False
    # Check if all values are zero in column
    elif (df[col] == 0).all(axis=0):
        print("Load at trafo {a} is zero, skipping column".format(a=col))
        return False
    # Check if all values are NaN in the column
    elif np.all(pd.isna(col)):
        print("Load at trafo {a} is missing, skipping column".format(a=col))
        return False
    return True


def calc_completeness_dataframe(
    df: pd.DataFrame,
    time_delayed: bool = False,
    homogenise: bool = True,
) -> pd.DataFrame:
    """Calculate the completeness of each column in dataframe.

    NOTE: NA values count as incomplete

    Args:
        df: Dataframe with a datetimeIndex index
        time_delayed: Should there be a correction for T-x columns
        homogenise: Should the index be resampled to median time delta -
            only available for DatetimeIndex

    Returns:
        Dataframe with fraction of completeness per column

    """
    logger = structlog.get_logger(__name__)

    if homogenise and isinstance(df.index, pd.DatetimeIndex) and len(df) > 0:

        median_timediff = int(
            df.reset_index().iloc[:, 0].diff().median().total_seconds() / 60.0
        )
        df = df.resample("{:d}T".format(median_timediff)).mean()

    if time_delayed is False:
        # Calculate completeness
        # Completeness per column
        completeness_per_column_dataframe = df.count() / len(df)

    # if timeDelayed is True, we correct that time-delayed columns
    # also in the best case will have NA values. E.g. T-2d is not available
    # for times ahead of more than 2 days
    elif time_delayed:
        # assume 15 minute forecast resolution
        # timecols: {delay:number of points expected to be missing}
        # number of points expected to be missing = numberOfPointsUpToTwoDaysAhead - numberOfPointsAvailable
        timecols = {
            x: len(df) - eval(x[2:].replace("min", "/60").replace("d", "*24.0")) / 0.25
            for x in df.columns
            if x[:2] == "T-"
        }

        non_na_count = df.count()

        for col, value in timecols.items():
            if non_na_count[col] > value:
                logger.warning(
                    "The provided input data (features) contains more values than is to be expected from analysis",
                    feature=col,
                    number_non_na=non_na_count[col],
                    expected_numbers_timedelayed=value,
                )

        # Correct for APX being only expected to be available up to 24h
        if "APX" in non_na_count.index:
            non_na_count["APX"] += max([len(df) - 96, 0])

        completeness_per_column_dataframe = non_na_count / (len(df))

    return completeness_per_column_dataframe



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from enum import Enum

import numpy as np
import pandas as pd
import structlog

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import ForecastType
from openstef.feature_engineering import weather_features

# this is the default for "Lagerwey100"
TURBINE_DATA = {
    "rated_power": 1,
    "slope_center": 8.07,
    "steepness": 0.664,
}

# Set value to define precission of power, this is needed because comparing to zero sometimes leads to issues for very small values.
SMALLEST_POWER_UNIT: float = 0.000001


def normalize_and_convert_weather_data_for_splitting(
    weather_data: pd.DataFrame,
) -> pd.DataFrame:
    """Normalize and converts weather data for use in energy splitting.

    Args:
        weather_data: Weather data with "windspeed_100m" and "radiation".

    Returns:
         Dataframe with "windpower" and "radiation" columns.

    """
    # Check we have "windspeed_100m" and "radiation" available
    if not all(
        elem in weather_data.columns for elem in ["windspeed_100m", "radiation"]
    ):
        raise ValueError("weather data does not contain required data!")

    # Prepare output dataframe
    output_dataframe = pd.DataFrame()

    # Normalize weather data
    output_dataframe["radiation"] = (
        weather_data["radiation"]
        / np.percentile(weather_data["radiation"].dropna(), 99.0)
        * -1
    )
    wind_ref_series = weather_features.calculate_windspeed_at_hubheight(
        weather_data["windspeed_100m"], fromheight=100
    )
    wind_ref = wind_ref_series.to_frame()
    wind_ref = calculate_wind_power(wind_ref)
    wind_ref *= -1

    output_dataframe["windpower"] = wind_ref
    return output_dataframe


def calculate_wind_power(
    windspeed_100m: pd.DataFrame,
) -> pd.DataFrame:
    """Calculate the generated wind power based on the wind speed.

    Values are related through the power curve, which is
    described by turbine_data. Default values are used and are normalized to 1MWp.

    Args:
        windspeed_100m: Example: ``pd.DataFrame (index = datetime, columns = ["windspeed_100m"])``

    Returns:
        Example output ``pd.DataFrame(index = datetime, columns = ["windenergy"])``

    """
    generated_power = TURBINE_DATA["rated_power"] / (
        1
        + np.exp(
            -TURBINE_DATA["steepness"] * (windspeed_100m - TURBINE_DATA["slope_center"])
        )
    )
    return generated_power["windspeed_100m"].rename("windenergy").to_frame()


def split_forecast_in_components(
    forecast: pd.DataFrame, weather_data: pd.DataFrame, split_coefs: dict
) -> dict[str, pd.DataFrame]:
    """Make estimates of energy components based on given forecast.

    Args:
        forecast: KTP load forecast
        weather_data: Weather data for energy splitting, at least; "windspeed_100m" and "radiation"
        split_coefs: Previously determined splitting coefs for prediction job

    Returns:
        Forecast dataframe for each component

    """
    # Normalize weather data
    weather_ref_profiles = normalize_and_convert_weather_data_for_splitting(
        weather_data
    )

    # Check input
    if not all(
        elem in ["windpower", "radiation"]
        for elem in list(weather_ref_profiles.columns)
    ):
        raise ValueError("weather data does not contain required data!")

    # Merge to ensure datetime index is the same
    weather_ref_profiles = forecast.merge(
        weather_ref_profiles, how="outer", right_index=True, left_index=True
    )
    # Drop rows with duplicate indices
    weather_ref_profiles = weather_ref_profiles[
        ~weather_ref_profiles.index.duplicated()
    ]
    weather_ref_profiles.replace([np.inf, -np.inf], np.nan).dropna(inplace=True)

    # Prepare output dictionary and list of forecast types
    components = forecast.copy(deep=True)

    # Calculate profiles of estimated components
    components["forecast_wind_on_shore"] = (
        split_coefs["wind_ref"] * weather_ref_profiles["windpower"]
    )
    components["forecast_solar"] = (
        split_coefs["pv_ref"] * weather_ref_profiles["radiation"]
    )
    components["forecast_other"] = (
        weather_ref_profiles["forecast"]
        - components["forecast_solar"]
        - components["forecast_wind_on_shore"]
    )

    # Check that sign of production components is negative and not positive, change if sign is wrong
    if components["forecast_wind_on_shore"].sum() > 0:
        raise ValueError("Sign of estimated wind_on_shore component is positive!")
    if components["forecast_solar"].sum() > 0:
        raise ValueError("Sign of estimated solar component is positive!")

    # Post process predictions to ensure realistic values
    components["forecast_solar"] = post_process_wind_solar(
        components["forecast_solar"], ForecastType.SOLAR
    )
    components["forecast_wind_on_shore"] = post_process_wind_solar(
        components["forecast_wind_on_shore"], ForecastType.WIND
    )

    return components.drop("forecast", axis=1).drop("stdev", axis=1).dropna()


def post_process_wind_solar(
    forecast: pd.Series, forecast_type: ForecastType
) -> pd.DataFrame:
    """Function that caries out postprocessing for wind and solar power generators.

        As these points will always produce energy, predicted energy consumption is
        set to zero. This function enforces the assumption that production is negative
        and consuption positive.

    Args:
        forecast: Series with forecast data.
        forecast_type: Specifies the type of forecast. This can be retrieved
            from the prediction job as pj['forecast_type']

    Returns:
        Post-processed forecast.

    """
    if forecast_type not in [ForecastType.WIND, ForecastType.SOLAR]:
        return forecast

    # For wind and solar forecasted value should always be negative.
    forecast.loc[forecast > (-1 * SMALLEST_POWER_UNIT)] = 0

    # write changed back to forecast
    return forecast


def add_components_base_case_forecast(basecase_forecast: pd.DataFrame) -> pd.DataFrame:
    """Makes a basecase forecast for the forecast_other component.

    This will make a simple basecase components forecast
    available and ensures that the sum of the components (other, wind and solar) is equal to the normal basecase
    forecast This is important for sending GLMD messages correctly to TenneT!

    Args:
        basecase_forecast: pd.DataFrame with basecase forecast

    Returns:
        basecase_forecast: pd.DataFrame with extra "forecast_other component"

    """
    basecase_forecast["forecast_other"] = basecase_forecast["forecast"]
    return basecase_forecast


def add_prediction_job_properties_to_forecast(
    pj: PredictionJobDataClass,
    forecast: pd.DataFrame,
    algorithm_type: str,
    forecast_type: Enum = None,
    forecast_quality: str = None,
) -> pd.DataFrame:
    """Adds prediciton job meta data to a forecast dataframe.

    Args:
        pj: Prediciton job.
        forecast: Forecast dataframe
        algorithm_type: Type of algirithm used for making the forecast.
        forecast_type: Type of the forecast. Defaults to None.
        forecast_quality: Quality of the forecast. Defaults to None.

    Returns:
        Dataframe with added metadata.

    """
    logger = structlog.get_logger(__name__)

    logger.info("Postproces in preparation of storing")
    if forecast_type is None:
        forecast_type = pj["forecast_type"]
    else:
        # get the value from the enum
        forecast_type = forecast_type.value

    # NOTE this field is only used when making the babasecase forecast and fallback
    if forecast_quality is not None:
        forecast["quality"] = forecast_quality

    # TODO rename prediction job typ to type
    # TODO algtype = model_file_path, perhaps we can find a more logical name
    # TODO perhaps better to make a forecast its own class!
    # TODO double check and sync this with make_basecase_forecast (other fields are added)
    # !!!!! TODO fix the requirement for customer
    forecast["pid"] = pj["id"]
    forecast["customer"] = pj["name"]
    forecast["description"] = pj["description"]
    forecast["type"] = forecast_type
    forecast["algtype"] = algorithm_type

    return forecast



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""Defines reporter class."""
import os
import warnings
from dataclasses import dataclass

import numpy as np
import pandas as pd
import sklearn
import structlog
from mlflow.models import ModelSignature, infer_signature
from plotly.graph_objects import Figure

from openstef.metrics import figure
from openstef.metrics.metrics import bias, mae, nsme, r_mae, rmse
from openstef.model.regressors.regressor import OpenstfRegressor


@dataclass
class Report:
    """Dataclass to hold a report describing the training process."""

    def __init__(
        self,
        feature_importance_figure: Figure,
        data_series_figures: dict[str, Figure],
        metrics: dict,
        signature: ModelSignature,
    ):
        """Initialize training report.

        Args:
            feature_importance_figure: Figure with feature importance
            data_series_figures: Figure with input data time series.
            metrics: Dict with metrics
            signature: Model signature

        """
        self.feature_importance_figure = feature_importance_figure
        self.data_series_figures = data_series_figures
        self.metrics = metrics
        self.signature = signature


class Reporter:
    """Reporter class that generates reports describing the training process."""

    def __init__(
        self,
        train_data: pd.DataFrame = None,
        validation_data: pd.DataFrame = None,
        test_data: pd.DataFrame = None,
    ) -> None:
        """Initializes reporter.

        Args:
            train_data: Dataframe with training data
            validation_data: Dataframe with validation data
            test_data: Dataframe with test data

        """
        self.horizons = train_data.horizon.unique()
        self.predicted_data_list = []
        self.input_data_list = [train_data, validation_data, test_data]

    def generate_report(
        self,
        model: OpenstfRegressor,
    ) -> Report:
        """Generate a report on a given model.

        Args:
            model: the model to create a report on

        Returns:
            Reporter object containing info about the model

        """
        # Get training (input_data_list[0]) and validation (input_data_list[1]) set
        train_x, train_y = (
            self.input_data_list[0].iloc[:, 1:-1],
            self.input_data_list[0].iloc[:, 0],
        )
        valid_x, valid_y = (
            self.input_data_list[1].iloc[:, 1:-1],
            self.input_data_list[1].iloc[:, 0],
        )

        data_series_figures = self._make_data_series_figures(model)

        # feature_importance_dataframe should be a dataframe, to create a figure
        # can be None if we have no feature importance
        if isinstance(model.feature_importance_dataframe, pd.DataFrame):
            feature_importance_figure = figure.plot_feature_importance(
                model.feature_importance_dataframe
            )
        # If it isn't a dataframe we will set feature_importance_figure, so it will not create the figure
        else:
            feature_importance_figure = None

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            report = Report(
                data_series_figures=data_series_figures,
                feature_importance_figure=feature_importance_figure,
                metrics=self.get_metrics(model.predict(valid_x), valid_y),
                signature=infer_signature(train_x, train_y),
            )

        return report

    @staticmethod
    def get_metrics(y_pred: np.array, y_true: np.array) -> dict:
        """Calculate the metrics for a prediction.

        Args:
            y_pred: np.array
            y_true: np.array

        Returns:
            Metrics for the prediction

        """
        metric_dict = {
            "bias": bias,
            "NSME": nsme,
            "MAE": mae,
            "R_MAE": r_mae,
            "RMSE": rmse,
            "explained_variance": sklearn.metrics.explained_variance_score,
            "MSE": sklearn.metrics.mean_squared_error,
            "r2": sklearn.metrics.r2_score,
        }
        results = {}
        for name, metric in metric_dict.items():
            try:
                results[name] = metric(y_true, y_pred)
            except ValueError:
                continue
        return results

    @staticmethod
    def write_report_to_disk(report: Report, report_folder: str):
        """Write report to disk; e.g. for viewing report of latest models using grafana."""
        # Initialize logger and serializer
        logger = structlog.get_logger(__name__)
        if report_folder:
            # create path if does not exist
            if not os.path.exists(report_folder):
                os.makedirs(report_folder)
            logger.info(f"Writing reports to {report_folder}")
            # write feature importance figure
            if report.feature_importance_figure:  # only write if figure is not none
                report.feature_importance_figure.write_html(
                    os.path.join(report_folder, "weight_plot.html")
                )
            # write predictors
            for name, figure in report.data_series_figures.items():
                if figure:  # only write if figure is not none
                    figure.write_html(os.path.join(report_folder, f"{name}.html"))

    def _make_data_series_figures(self, model: OpenstfRegressor) -> dict:
        """Make data series figures."""
        # Make model predictions
        for data_set in self.input_data_list:
            # First ("load") and last ("horizon") are removed here
            # as they are not expected by the model as prediction input
            model_forecast = model.predict(data_set.iloc[:, 1:-1])
            forecast = pd.DataFrame(
                index=data_set.index, data={"forecast": model_forecast}
            )
            self.predicted_data_list.append(forecast)

        # Make cufflinks plots for the data series
        return {
            f"Predictor{horizon}": figure.plot_data_series(
                data=self.input_data_list,
                predict_data=self.predicted_data_list,
                horizon=horizon,
            )
            for horizon in self.horizons
        }



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import random
import secrets
from datetime import timedelta
from itertools import accumulate
from typing import Iterable

import numpy as np
import pandas as pd

AMOUNT_DAY = 96  # Duration of the periods (in T-15) that are in a day (default = 96)
PERIOD_TIMEDELTA = 1  # Duration of the periods (in days) that will be sampled as validation data for each split.
PEAK_FRACTION = 0.15


def group_kfold(
    input_data: pd.DataFrame, n_folds: int, randomize_fold_split: bool = True
) -> pd.DataFrame:
    """Function to group data into groups, according to the date and the number of folds.

    Each date gets assigned a number between 0 and n_folds.

    Args:
        input_data: Input data
        n_folds: Number of folds
        randomize_fold_split: Indicates if random split needs to be applied

    Returns:
        Grouped data

    """
    unique_dates = input_data["dates"].unique()  # dates defines the day (Y-M-D)
    # Group separators
    len_data = len(unique_dates)  # number of indices that can be used for splitting
    size = len_data // n_folds  # size of each fold set
    rem = (
        len_data % n_folds
    )  # remaining number of indices when divided in fold sets of equal size
    separators = list(
        accumulate([0] + [size + 1] * rem + [size] * (n_folds - rem))
    )  # location of seperators

    items = list(unique_dates)

    if randomize_fold_split:
        random.shuffle(items)  # if random, shuffle the days

    for i, s in enumerate(zip(separators, separators[1:])):
        group = items[slice(*s)]
        input_data.loc[
            input_data[input_data["dates"].isin(group)].index, "random_fold"
        ] = i
    return input_data


def sample_indices_train_val(
    data: pd.DataFrame, peaks: pd.DataFrame
) -> tuple[np.array, np.array]:
    """Sample indices of given period length assuming the peaks are evenly spreaded.

    Args:
        data: Clean data with features
        peaks: Data frame of selected peaks to sample the dates from

    Returns:
        - List with the start point of each peak
        - Sorted list with the indices corresponding to the peak

    """
    sampled = set()
    peaks_val = []

    for peak in peaks:
        sampled |= set(data[data.index.date == peak].index)
        peaks_val.append(peak)
    return peaks_val, np.sort(list(sampled))


def random_sample(all_peaks: np.array, k: int) -> np.array:
    """Random sampling of numbers out of a np.array.

    Implemented due to security sonar cloud not accepting the random built-in functions.

    Args:
        all_peaks: List with numbers to sample from
        k: Number of wanted samples

    Returns:
        Sorted array with the random samples (dates from the peaks)

    """
    random_peaks = []
    all_peaks_list = all_peaks.tolist()
    for _ in range(k):
        element_random = secrets.choice(all_peaks_list)
        all_peaks_list.remove(element_random)
        random_peaks.append(element_random)
    return np.array(random_peaks)


def split_data_train_validation_test(
    data_: pd.DataFrame,
    test_fraction: float = 0.1,
    validation_fraction: float = 0.15,
    back_test: bool = False,
    stratification_min_max: bool = True,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Split input data into train, test and validation set.

    Function for splitting data with features in a train, test and
    validation dataset. In an operational setting the following sequence is
    returned (when using stratification):

    Test >> Train >> Validation

    For a back test (indicated with argument "back_test") the following sequence
    is returned:

    Train >> Validation >> Test

    The ratios of the different types can be set with test_fraction and
    validation fraction.

    Args:
        data_: Cleaned data with features
        test_fraction : Number between 0 and 1 that indicates the desired
            fraction of test data.
        validation_fraction: Number between 0 and 1 that indicates the
            desired fraction of validation data.
        back_test: Indicates if data is intended for a back test.
        stratification_min_max: Indicates if validation data must be sampled as
            periods, using stratification on min and max values per day.
            If True, 'extreme days' are ensured to be included in the validation and train sets,
            ensuring the validation set to be representative of the train set.

    Returns:
        - Train data.
        - Validation data.
        - Test data.

    """
    train_fraction = 1 - (test_fraction + validation_fraction)
    if train_fraction < 0:
        raise ValueError(
            "Test ({test_fraction}) and validation fraction ({validation_fraction}) too"
            " high."
        )

    # Define constants
    min_days_for_stratification = 4
    extreme_values_fraction = (
        0.15  # 15percent of highest and lowest values are considered extremes
    )

    # Get start date from the index
    start_date = data_.index.min().to_pydatetime()
    end_date = data_.index.max().to_pydatetime()

    # Calculate total of quarter hours (PTU's) in input data
    number_indices = len(data_.index.unique())  # Total number of unique timepoints
    delta = (
        data_.index.unique().sort_values()[1] - data_.index.unique().sort_values()[0]
    )  # Delta t, assumed to be constant throughout DataFrame
    delta = timedelta(
        seconds=delta.seconds
    )  # Convert from pandas timedelta to original python timedelta

    # Determine which dates are in testset
    if back_test:
        start_date_test = end_date - np.round(number_indices * test_fraction) * delta
        test_data = data_[start_date_test:]
        train_val_data = data_[:start_date_test]
    else:
        start_date_val = start_date + np.round(number_indices * test_fraction) * delta
        test_data = data_[:start_date_val]
        train_val_data = data_[start_date_val:]

    if stratification_min_max and (
        len(set(train_val_data.index.date)) >= min_days_for_stratification
    ):
        # First how many dates are considers min or max.
        # Note that this should be at least to, so one can go to the train set
        # and another to the testset
        train_val_dates = list(set(train_val_data.index.date))
        n_days_per_min_max_subset = int(
            max(extreme_values_fraction * len(set(data_.index.date)), 2)
        )
        # Find max_dates
        max_dates = (
            train_val_data[["load"]]
            .resample("1D")
            .max()
            .sort_values(by="load", ascending=False)
            .dropna()
            .index[:n_days_per_min_max_subset]
        ).date
        # Find min_dates, but do not consider the max_dates
        min_dates_subset = train_val_data.loc[
            ~np.isin(train_val_data.index.date, max_dates), ["load"]
        ]
        min_dates = (
            min_dates_subset[["load"]]
            .resample("1D")
            .min()
            .sort_values(by="load", ascending=True)
            .dropna()
            .index[:n_days_per_min_max_subset]
        ).date
        other_dates = [
            x for x in train_val_dates if x not in min_dates and x not in max_dates
        ]

        # Divide min, max and other dates fairly over validation and train set, with at least 1 min and max in train and validation
        val_dates = []
        train_dates = []
        for date_set in [max_dates, min_dates, other_dates]:
            n_days_val = max(1, int(validation_fraction * len(date_set)))
            val_dates += list(
                np.random.choice(list(date_set), n_days_val, replace=False)
            )
            train_dates += [x for x in date_set if x not in val_dates]

        validation_data = train_val_data[np.isin(train_val_data.index.date, val_dates)]
        train_data = train_val_data[np.isin(train_val_data.index.date, train_dates)]

    # Default sampling, take a one single validation set.
    else:
        if back_test:
            start_date_train = (
                start_date + np.round(number_indices * validation_fraction) * delta
            )
            end_date_train = end_date - np.round(number_indices * test_fraction) * delta
            validation_data = data_[:start_date_train]
            train_data = data_[start_date_train:end_date_train]
        else:
            start_date_val = (
                start_date + np.round(number_indices * test_fraction) * delta
            )
            start_date_train = (
                start_date_val + np.round(number_indices * validation_fraction) * delta
            )
            train_data = data_[start_date_train:None]
            validation_data = data_[start_date_val:start_date_train]

    train_data = train_data.sort_index()
    validation_data = validation_data.sort_index()
    test_data = test_data.sort_index()

    return (
        train_data,
        validation_data,
        test_data,
    )


def backtest_split_default(
    data: pd.DataFrame,
    n_folds: int,
    test_fraction: float = 0.15,
    stratification_min_max: bool = True,
    randomize_fold_split: bool = False,
) -> Iterable[tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """Default cross validation strategy.

    Args:
        data:
        n_folds:
        test_fraction:
        stratification_min_max:
        randomize_fold_split:

    Returns:
        Iterable on train, val, test splits

    Notes:
        We use a generator in order to have lazy estimation and avoid multiple copy of the data.

    """
    if n_folds > 1:
        data.index = pd.to_datetime(data.index)
        data["dates"] = data.index
        data = group_kfold(data, n_folds, randomize_fold_split)

        for ifold in range(n_folds):
            test_data = data[data["random_fold"] == ifold].sort_index()

            (train_data, validation_data, _,) = split_data_train_validation_test(
                data[data["random_fold"] != ifold].iloc[:, :-2],
                test_fraction=0,
                back_test=True,
                stratification_min_max=stratification_min_max,
            )

            yield train_data, validation_data, test_data.iloc[:, :-2]
    else:
        yield split_data_train_validation_test(
            data,
            back_test=True,
            test_fraction=test_fraction,
            stratification_min_max=stratification_min_max,
        )



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

from setuptools import find_packages, setup

pkg_dir = Path(__file__).parent.absolute()


def read_requirements_from_file():
    with open(pkg_dir / "requirements.txt") as fh:
        requirements = []
        for line in fh:
            line = line.strip()
            if "#" in line:
                line = line[: line.index("#")].strip()
            if len(line) == 0:
                continue
            requirements.append(line)
        return requirements


def read_long_description_from_readme():
    with open("README.md", "r", encoding="utf-8") as fh:
        return fh.read()


setup(
    name="openstef",
    version="3.2.64",
    packages=find_packages(include=["openstef", "openstef.*"]),
    description="Open short term energy forecaster",
    long_description=read_long_description_from_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/OpenSTEF/openstef",
    author="Alliander N.V",
    author_email="korte.termijn.prognoses@alliander.com",
    license="MPL-2.0",
    keywords=["energy", "forecasting", "machinelearning"],
    # See https://setuptools.readthedocs.io/en/latest/userguide/datafiles.html
    # for more information
    package_data={
        # Include anything in the data directory
        "openstef": ["data/*", "*.license"]
    },
    python_requires=">=3.9.0",
    install_requires=read_requirements_from_file(),
    setup_requires=["wheel"],
    tests_require=["pytest", "pytest-cov", "flake8"],
    extras_require={
        "proloaf": ["proloaf==0.2.0", "torch==1.10.0", "pytorch-lightning==1.5.1"]
    },
    classifiers=[
        r"Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        r"License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from datetime import datetime

import pandas as pd


def generate_fallback(
    forecast_input: pd.DataFrame,
    load: pd.DataFrame,
    fallback_strategy: str = "extreme_day",
) -> pd.DataFrame:
    """Make a fall back forecast, Set the value of the forecast 'quality' column to 'substituted'.

    Currently only fallback_strategy=extreme day is implemented which return historic profile of most extreme day.

    Args:
        forecast_input : dataframe desired for the forecast
        load: index=datetime, columns=['load']
        fallback_strategy: strategy to determine fallback. options:
            - extreme_day: use daily profile of most extreme day
    Returns:
        Fallback forecast DataFrame with columns; 'forecast', 'quality'

    Raises:
        ValueError if len(load) == 0
        NotImplementedError if fallback_strategy != 'extreme_day'

    """
    # Check if load is completely empty
    if len(load.dropna()) == 0:
        raise ValueError("No historic load data available")

    if fallback_strategy != "extreme_day":
        raise NotImplementedError(
            f'fallback_strategy should be "extreme_day", received:{fallback_strategy}'
        )

    if fallback_strategy == "extreme_day":
        # Execute this fallback strategy
        # Find most extreme historic day and merge it by time-of-day to the requested moments

        # Find most extreme historic day (do not count today as it is incomplete)
        day_with_highest_load_date = (
            load[load.index.tz_localize(None).date != datetime.utcnow().date()]
            .idxmax()
            .load.date()
        )
        # generate datetime range of the day with the highest load
        from_datetime = pd.Timestamp(day_with_highest_load_date, tz=load.index.tz)
        till_datetime = from_datetime + pd.Timedelta("1 days")

        # slice load dataframe, only rows for the day with the highest load
        highest_daily_loadprofile = load.loc[
            (load.index >= from_datetime) & (load.index < till_datetime)
        ]

        # Match moments by time-of-day
        highest_daily_loadprofile.loc[:, "time"] = highest_daily_loadprofile.index.time
        forecast = pd.DataFrame(index=forecast_input.index)
        forecast["time"] = forecast.index.time
        forecast = (
            forecast.reset_index()
            .merge(
                highest_daily_loadprofile, left_on="time", right_on="time", how="outer"
            )
            .set_index("index")
        )

    # Rename so column is called forecast
    forecast = forecast[["load"]].rename(columns=dict(load="forecast"))

    # Add a column quality.
    forecast["quality"] = "substituted"

    return forecast



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""Specifies the split function dataclass."""
import inspect
import json
from importlib import import_module
from typing import Any, Callable, Sequence, Union

from pydantic import BaseModel


class SplitFuncDataClass(BaseModel):
    """Class that allows to specify a custom function to generate a train, test and validation set."""

    function: Union[str, Callable]
    arguments: Union[
        str, dict[str, Any]
    ]  # JSON string holding the function parameters or dict

    def __getitem__(self, key: str):
        """Allows us to use subscription to get the items from the object."""
        return getattr(self, key)

    def __setitem__(self, key: str, value: any):
        """Allows us to use subscription to set the items in the object."""
        if hasattr(self, key):
            self.__dict__[key] = value
        else:
            raise AttributeError(f"{key} not an attribute of prediction job.")

    def _load_split_function(
        self, required_arguments: Sequence[str] = None
    ) -> Callable:
        """Load split function from path.

        Args:
            func_path (str): The path to the split function

        Returns:
            split_func (Callable): The loaded split function

        """
        if isinstance(self.function, str):
            path_elements = self.function.split(".")
            module_path = ".".join(path_elements[:-1])
            module = import_module(module_path)
            func_name = path_elements[-1]
            split_func = getattr(module, func_name)
        else:
            split_func = self.function

        # Check that the function accepts mandatory arguments
        if not callable(split_func):
            raise ValueError("The loaded object is not callable: {func_path!r}")

        if required_arguments is not None:
            func_params = set(inspect.signature(split_func).parameters)

            if len(set(required_arguments) - func_params) > 0:
                raise ValueError(
                    "The loaded split function does not have the required arguments"
                )

        return split_func

    def _load_arguments(self) -> dict[str, Any]:
        """Load the arguments.

        Convert the arguments from JSON if they are given as strings or simply return them otherwise.

        Returns:
            arguments (dict[str, Any]): The additional arguments to be passed to he function

        """
        if isinstance(self.arguments, str):
            return json.loads(self.arguments)
        else:
            return self.arguments

    def load(
        self, required_arguments: Sequence[str] = None
    ) -> tuple[Callable, dict[str, Any]]:
        """Load the function and its arguments.

        If the function and the arguments are given as strings in the instane attributes, load them as Python objects
        otherwise just return them from the instance attributes.

        Args:
            required_arguments (list[str]): list of arguments the loaded function must have

        Returns:
            - function (Callable)
            - arguments (dict[str, Any])

        """
        return self._load_split_function(required_arguments), self._load_arguments()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from xgboost import XGBRegressor

from openstef.model.regressors.regressor import OpenstfRegressor


class XGBOpenstfRegressor(XGBRegressor, OpenstfRegressor):
    """XGB Regressor which implements the Openstf regressor API."""

    gain_importance_name = "total_gain"
    weight_importance_name = "weight"

    @property
    def feature_names(self):
        return self._Booster.feature_names

    @property
    def can_predict_quantiles(self):
        return False

    @staticmethod
    def _get_importance_names():
        return {
            "gain_importance_name": "total_gain",
            "weight_importance_name": "weight",
        }



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This modelu contains various helper functions."""

import numpy as np
import pandas as pd
import structlog


def add_missing_feature_columns(
    input_data: pd.DataFrame, features: list[str]
) -> pd.DataFrame:
    """Adds feature column for features in the featurelist.

    Add feature columns for features in the feature list if these columns don't
    exist in the input data. If a column is added, its value is set to NaN.
    This is especially usefull to make sure the required columns are in place when
    making a prediction.

    .. note::
        This function is intended as a final check to prevent errors during predicion.
        In an ideal world this function is not nescarry.

    Args:
        input_data: DataFrame with input data and featurs.
        features: List of requiered features.

    Returns:
        Input dataframe with missing columns filled with ``np.N=nan``.

    """
    logger = structlog.get_logger(__name__)

    if features is None:
        features = []

    missing_features = [f for f in features if f not in list(input_data)]

    for feature in missing_features:
        logger.warning(
            f"Adding NaN column for missing feature: {feature}", missing_feature=feature
        )
        input_data[feature] = np.nan

    return input_data


def remove_non_requested_feature_columns(
    input_data: pd.DataFrame, requested_features: list[str]
) -> pd.DataFrame:
    """Removes features that are provided in the input data but not in the feature list.

    This should not be nescesarry but serves as an extra failsave for making predicitons

    Args:
        input_data: DataFrame with features
        requested_features: List of reuqested features

    Returns:
        Model input data with features.

    """
    logger = structlog.get_logger(__name__)

    if requested_features is None:
        requested_features = []

    not_requested_features = [
        f for f in list(input_data) if f not in requested_features
    ]

    # Do not see "load" or "horizon" as an extra feature as it is no feature
    if "load" in not_requested_features:
        not_requested_features.remove("load")

    num_not_requested_features = len(not_requested_features)

    if num_not_requested_features != 0:
        logger.warning(
            f"Removing {num_not_requested_features} unrequested features!",
            num_not_requested_features=num_not_requested_features,
        )

    return input_data.drop(not_requested_features, axis=1)


def enforce_feature_order(input_data: pd.DataFrame) -> pd.DataFrame:
    """Enforces correct order of features.

    Alphabetically orders the feature columns. The load column remains the first column
    and the horizons column remains the last column.
    Everything in between is alphabetically sorted:
    The order eventually looks like this:
    ["load"] -- [alphabetically sorted features] -- ['horizon']

    This function assumes the first column contains the to be predicted variable
    Furthermore the "horizon" is moved to the last position if it is pressent.

    Args:
        input_data: Input data with features.

    Returns:
        Properly sorted input data.

    """
    # Extract first column name
    first_column_name = input_data.columns.to_list()[
        0
    ]  # Most of the time this is "load"

    # Sort columns
    columns = list(np.sort(input_data.columns.to_list()))

    # Remove first column and add to the start
    columns.remove(first_column_name)
    column_order = [first_column_name] + columns

    # If "Horzion" column is available add to the end
    if "horizon" in columns:
        # "horizon" is pressent in the training procces
        # but not in the forecasting process
        column_order.remove("horizon")
        column_order = column_order + ["horizon"]

    # Return dataframe with columns in the correct order
    return input_data.loc[:, column_order]



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module provides functionality for applying features to the input data to improve forecast accuracy.

Examples of features that are added:
    - The load 1 day and 7 days ago at the same time.
    - If a day is a weekday or a holiday.
    - The extrapolated windspeed at 100m.
    - The normalised wind power according to the turbine-specific power curve.

"""

import pandas as pd

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.feature_engineering.historic_features import (
    add_historic_load_as_a_feature,
)
from openstef.feature_engineering.holiday_features import (
    generate_holiday_feature_functions,
)
from openstef.feature_engineering.lag_features import generate_lag_feature_functions
from openstef.feature_engineering.weather_features import (
    add_additional_solar_features,
    add_additional_wind_features,
    add_humidity_features,
)


def apply_features(
    data: pd.DataFrame,
    pj: PredictionJobDataClass = None,
    feature_names: list[str] = None,
    horizon: float = 24.0,
) -> pd.DataFrame:
    """Applies the feature functions defined in ``feature_functions.py`` and returns the complete dataframe.

    Features requiring more recent label-data are omitted.

    .. note::
        For the time deriven features only the onces in the features list will be added. But for the weather features all will be added at present.
        These unrequested additional features have to be filtered out later.

    Args:
        data (pandas.DataFrame): a pandas dataframe with input data in the form:
                                    pd.DataFrame(
                                        index=datetime,
                                        columns=[label, predictor_1,..., predictor_n]
                                    )
        pj (PredictionJobDataClass): Prediction job.
        feature_names (list[str]): list of reuqested features
        horizon (float): Forecast horizon limit in hours.

    Returns:
        pd.DataFrame(index = datetime, columns = [label, predictor_1,..., predictor_n, feature_1, ..., feature_m])

    Example output:

    .. code-block:: py

        import pandas as pd
        import numpy as np
        index = pd.date_range(start = "2017-01-01 09:00:00",
        freq = '15T', periods = 200)
        data = pd.DataFrame(index = index,
                            data = dict(load=
                            np.sin(index.hour/24*np.pi)*
                            np.random.uniform(0.7,1.7, 200)))

    """
    # Add if needed the proloaf feature (historic_load)
    data = add_historic_load_as_a_feature(data, pj)

    # Get lag feature functions
    feature_functions = generate_lag_feature_functions(feature_names, horizon)

    # Get timedrivenfeature functions
    feature_functions.update(
        {
            "IsWeekendDay": lambda x: (x.index.weekday // 5) == 1,
            "IsWeekDay": lambda x: x.index.weekday < 5,
            "IsSunday": lambda x: x.index.weekday == 6,
            "Month": lambda x: x.index.month,
            "Quarter": lambda x: x.index.quarter,
        }
    )

    # Get holiday feature functions
    feature_functions.update(generate_holiday_feature_functions())

    # Add the features to the dataframe using previously defined feature functions
    for key, featfunc in feature_functions.items():
        # Don't generate feature is not in features
        if feature_names is not None and key not in feature_names:
            continue
        data.loc[:, key] = data.iloc[:, [0]].apply(featfunc)

    # Add additional wind features
    data = add_additional_wind_features(data, feature_names)

    # Add humidity features
    data = add_humidity_features(data, feature_names)

    # Add solar features; when pj is unavailable a default location is used.
    data = add_additional_solar_features(data, pj, feature_names)

    # Return dataframe including all requested features
    return data



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd


def replace_repeated_values_with_nan(
    df: pd.DataFrame, max_length: int, column_name: str
) -> pd.DataFrame:
    """Replace sequentially repeated values with NaN.

    Args:
        df: Data with potential repeating values.
        max_length: Maximum length of sequence. Above are set to NaN.
        column_name: Column name of input dataframe with repeating values.

    Returns:
        DataFrame, similar to df, with the desired values set to NaN.

    """
    data = df.copy(deep=True)
    sequentials = data[column_name].diff().ne(0).cumsum()
    grouped_sequentials_over_max = sequentials.groupby(sequentials).head(max_length)
    data.loc[~data.index.isin(grouped_sequentials_over_max.index), column_name] = np.nan
    return data



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

# -*- coding: utf-8 -*-
"""optimize_hyper_params.py.

This module contains the CRON job that is periodically executed to optimize the
hyperparameters for the prognosis models.

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python optimize_hyperparameters.py

"""
from datetime import datetime, timedelta
from pathlib import Path

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import MLModelType, PipelineType
from openstef.model.serializer import MLflowSerializer
from openstef.monitoring import teams
from openstef.pipeline.optimize_hyperparameters import optimize_hyperparameters_pipeline
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

MAX_AGE_HYPER_PARAMS_DAYS = 31
DEFAULT_CHECK_HYPER_PARAMS_AGE = True
DEFAULT_TRAINING_PERIOD_DAYS = 121


def optimize_hyperparameters_task(
    pj: PredictionJobDataClass,
    context: TaskContext,
    check_hyper_param_age: bool = DEFAULT_CHECK_HYPER_PARAMS_AGE,
) -> None:
    """Optimize hyperparameters task.

    Expected prediction job keys: "id", "model", "lat", "lon", "name", "description"
    Only used for logging: "name", "description"

    Args:
        pj: Prediction job
        context: Task context
        check_hyper_param_age: Boolean indicating if optimization can be skipped in case existing
            hyperparameters do not exceed the maximum age.

    """
    # Check pipeline types
    if PipelineType.HYPER_PARMATERS not in pj.pipelines_to_run:
        context.logger.info(
            "Skip this PredictionJob because hyper_parameters pipeline is not specified in the pj."
        )
        return

    # Retrieve the paths for storing model and reports from the config manager
    mlflow_tracking_uri = context.config.paths_mlflow_tracking_uri
    artifact_folder = context.config.paths_artifact_folder

    # Determine if we need to optimize hyperparams
    # retrieve last model age where hyperparameters were optimized
    mlflow_serializer = MLflowSerializer(mlflow_tracking_uri=mlflow_tracking_uri)
    hyper_params_age = mlflow_serializer.get_model_age(
        experiment_name=str(pj["id"]), hyperparameter_optimization_only=True
    )

    if (hyper_params_age < MAX_AGE_HYPER_PARAMS_DAYS) and check_hyper_param_age:
        context.logger.warning(
            "Skip hyperparameter optimization",
            pid=pj["id"],
            hyper_params_age=hyper_params_age,
            max_age=MAX_AGE_HYPER_PARAMS_DAYS,
        )
        return

    datetime_start = datetime.utcnow() - timedelta(days=DEFAULT_TRAINING_PERIOD_DAYS)
    datetime_end = datetime.utcnow()

    input_data = context.database.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    # Optimize hyperparams
    hyperparameters = optimize_hyperparameters_pipeline(
        pj,
        input_data,
        mlflow_tracking_uri=mlflow_tracking_uri,
        artifact_folder=artifact_folder,
    )

    # Sent message to Teams
    title = (
        f'Optimized hyperparameters for prediction job {pj["name"]} {pj["description"]}'
    )

    teams.post_teams(teams.format_message(title=title, params=hyperparameters))


def main(config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specify a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        model_type = [ml.value for ml in MLModelType]

        PredictionJobLoop(context, model_type=model_type).map(
            optimize_hyperparameters_task, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Union

import pandas as pd
import pymsteams
import structlog
from pymsteams import cardsection


def post_teams(
    msg: Union[str, dict],
    invalid_coefficients: pd.DataFrame = None,
    coefficients_df: pd.DataFrame = None,
    url: str = None,
    proxies: dict = None,
) -> None:
    """Post a message to Teams - KTP.

    Note that currently no authentication occurs.
    Security is given by keeping the URL secret.
    One should therefore refrain from using more enhanced features such as
    action buttons.

    Args:
        msg: For simple messages a string can be passed. For more
            complex messages pass a dict. The following keys are supported:
            text, links, sections. Each section can contain the following keys:
            text, title, images, facts, markdown. Also see:
            https://docs.microsoft.com/en-us/outlook/actionable-messages/send-via-connectors
        invalid_coefficients: df of information of invalid
            coefficients. Defaults to None.
        coefficients_df: df of new coefficients. Defaults to None.
        url: webhook url, monitoring by default
        proxies: Optinonal proxy settings.

    Note:
        This function is namespace-specific.
    """
    logger = structlog.get_logger(__name__)
    # If no url is passed, give warning and don't send teams message
    if url is None:
        logger.warning("Can't post Teams message, no url given.")
        return

    # Add invalid coefficients and manual coefficients-query to message
    if invalid_coefficients is not None and coefficients_df is not None:
        # add invalid coefficient information to message in dict-format
        invalid_coefficients_text = "".join(
            [
                f"\n* **{row.coef_name}**: {round(row.coef_value_new, 2)}, "
                f"(previous: {round(row.coef_value_last, 2)})"
                for index, row in invalid_coefficients.iterrows()
            ]
        )
        query = build_sql_query_string(coefficients_df, "energy_split_coefs")
        query_text = (
            "If you would like to update the coefficients manually in the "
            + "database, use this query:"
        )
        msg = {
            "fallback": msg,
            "title": "Invalid energy splitting coefficients",
            "text": msg,
            "sections": [
                {
                    "text": invalid_coefficients_text,
                    "markdown": True,
                },
                {
                    "title": "Manual query",
                    "text": query_text,
                    "markdown": True,
                },
                {
                    "text": query,
                    "markdown": True,
                },
            ],
        }

    card = pymsteams.connectorcard(url)

    # add proxies
    # NOTE the connectorcard.proxy is passed to the requests library under the hood
    card.proxies = proxies

    # if msg is string, convert to dict
    if type(msg) is str:
        msg = dict(text=msg)
    card.text(msg.get("text"))
    card.summary(msg.get("fallback", "-"))

    # set title, color, ...
    card.color(msg.get("color", "white"))
    card.title(msg.get("title"))

    link_dicts = msg.get("links", [])  # link_dicts can be single dict or list of dicts
    if isinstance(link_dicts, dict):  # if single dict
        card.addLinkButton(link_dicts["buttontext"], link_dicts["buttonurl"])
    elif isinstance(link_dicts, list):  # if list of dicts
        for link_dict in link_dicts:
            card.addLinkButton(link_dict["buttontext"], link_dict["buttonurl"])

    # Add sections
    for section_dict in msg.get("sections", []):
        card_section = get_card_section(section_dict=section_dict)
        card.addSection(card_section)

    card.send()


def get_card_section(section_dict: dict) -> cardsection:
    """Get card section for teams message from dictionary."""
    card_section = cardsection()
    card_section.text(section_dict.get("text"))
    card_section.title(section_dict.get("title"))
    for image in section_dict.get("images", []):
        card_section.addImage(image)
    for fact in section_dict.get("facts", []):
        card_section.addFact(*fact)
    if not section_dict.get("markdown", True):
        card_section.disableMarkdown()
    if "link" in section_dict:
        card_section.linkButton(
            section_dict.get("link").get("buttontext"),
            section_dict.get("link").get("buttonurl"),
        )
    return card_section


def build_sql_query_string(df: pd.DataFrame, table: str) -> str:
    """Build sql insert query string for Teams message output from df.

    Args:
        df: Df of table values to insert in sql.
        table: Table to insert df into.

    Returns:
        Sql query of insert statement.

    """
    # round all values to two decimals
    df = df.round(2)
    # convert datetime to string format
    datetime_columns = df.columns[
        df.columns.isin(["date_start", "date_end", "created"])
    ]
    for col in datetime_columns:
        df[col] = df[col].astype("str")

    sql_texts = [
        "```  \nINSERT INTO "
        + table
        + " ("
        + str(", ".join(df.columns))
        + ") VALUES  \n"
    ]
    for index, row in df.iterrows():
        if index != df.index[0]:
            sql_texts.append(",  \n")  # 2 spaces and \n create a new line
        sql_texts.append(str(tuple(row.values)))
    sql_texts.append("  \n```")
    query = "".join(sql_texts)
    return query


def format_message(title: str, params: dict, fallback=None, color=None) -> dict:
    if color is None:
        color = "#046b00"  # green
    if fallback is None:
        fallback = title

    # format allparams using limited precision for floats
    # make keys bold (**key**)
    values = []
    for k, v in params.items():
        if type(v) is float:
            values.append(f"**{k}**: {v:0.3f}")
            continue
        values.append(f"**{k}**: {v}")
    # join all {params}: {value}  pairs with a new line
    text = "".join([f"* {v}   \n" for v in values])

    msg = {
        "fallback": fallback,
        "title": title,
        "text": text,
        "color": color,
    }
    return msg



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from functools import partial

import numpy as np
import xgboost as xgb
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from xgboost import Booster

import openstef.metrics.metrics as metrics
from openstef.model.regressors.regressor import OpenstfRegressor

DEFAULT_QUANTILES: tuple[float, ...] = (0.9, 0.5, 0.1)


class XGBQuantileOpenstfRegressor(OpenstfRegressor):
    @staticmethod
    def _get_importance_names():
        return {
            "gain_importance_name": "total_gain",
            "weight_importance_name": "weight",
        }

    def __init__(
        self,
        quantiles: tuple[float, ...] = DEFAULT_QUANTILES,
        gamma: float = 0.0,
        colsample_bytree: float = 1.0,
        subsample: float = 1.0,
        min_child_weight: int = 1,
        max_depth: int = 6,
        learning_rate: float = 0.3,
        alpha: float = 0.0,
        max_delta_step: int = 0,
    ):
        """Initialize XGBQuantileRegressor.

        Model that provides quantile regression with XGBoost.
        For each desired quantile an XGBoost model is trained,
        these can later be used to predict quantiles.

        Args:
            quantiles: Tuple with desired quantiles, quantile 0.5 is required.
                For example: (0.1, 0.5, 0.9)
            gamma: Gamma
            colsample_bytree: Colsample by tree
            subsample: Subsample
            min_child_weight: Minimum child weight
            max_depth: Maximum depth
            learning_rate: Learning rate
            alpha: Alpha
            max_delta_step: Maximum delta step

        """
        super().__init__()
        # Check if quantile 0.5 is pressent this is required
        if 0.5 not in quantiles:
            raise ValueError(
                "Cannot train quantile model as 0.5 is not in requested quantiles!"
            )

        self.quantiles = quantiles

        # Set attributes for hyper parameters
        self.subsample = subsample
        self.min_child_weight = min_child_weight
        self.max_depth = max_depth
        self.gamma = gamma
        self.alpha = alpha
        self.max_delta_step = max_delta_step
        self.colsample_bytree = colsample_bytree
        self.learning_rate = learning_rate

    def fit(self, x: np.array, y: np.array, **kwargs) -> OpenstfRegressor:
        """Fits xgb quantile model.

        Args:
            x: Feature matrix
            y: Labels

        Returns:
            Fitted XGBQuantile model

        """
        # TODO: specify these required kwargs in the function definition
        early_stopping_rounds = kwargs.get("early_stopping_rounds", None)
        eval_set = kwargs.get("eval_set", None)

        # Check/validate input
        check_X_y(x, y, force_all_finite="allow-nan")

        # Convert x and y to dmatrix input
        dtrain = xgb.DMatrix(x.copy(deep=True), label=y.copy(deep=True))

        # Define watchlist if eval_set is defined
        if eval_set:
            dval = xgb.DMatrix(
                eval_set[1][0].copy(deep=True),
                label=eval_set[1][1].copy(deep=True),
            )

            # Define data set to be monitored during training, the last(validation)
            #  will be used for early stopping
            watchlist = [(dtrain, "train"), (dval, "validation")]
        else:
            watchlist = ()

        # Get fitting parameters - only those required for xgbooster's
        xgb_regressor_params = {
            key: value
            for key, value in self.get_params().items()
            if key in xgb.XGBRegressor().get_params().keys()
        }

        quantile_models = {}

        for quantile in self.quantiles:
            # Define objective callback functions specifically for desired quantile
            xgb_quantile_eval_this_quantile = partial(
                metrics.xgb_quantile_eval, quantile=quantile
            )
            xgb_quantile_obj_this_quantile = partial(
                metrics.xgb_quantile_obj, quantile=quantile
            )

            # Train quantile model
            quantile_models[quantile] = xgb.train(
                params=xgb_regressor_params,
                dtrain=dtrain,
                evals=watchlist,
                # Can be large because we are early stopping anyway
                num_boost_round=100,
                obj=xgb_quantile_obj_this_quantile,
                feval=xgb_quantile_eval_this_quantile,
                verbose_eval=False,
                early_stopping_rounds=early_stopping_rounds,
            )

        # Set weigths and features from the 0.5 (median) model
        self.feature_importances_ = self.get_feature_importances_from_booster(
            quantile_models[0.5]
        )
        self._Booster = quantile_models[0.5]  # Used for feature names later on
        # Update state of the estimator
        self.estimators_ = quantile_models
        self.is_fitted_ = True

        return self

    def predict(self, x: np.array, quantile: float = 0.5) -> np.array:
        """Makes a prediction for a desired quantile.

        Args:
            x: Feature matrix
            quantile: Quantile for which a prediciton is desired,
                note that only quantile are available for which a model is trained,
                and that this is a quantile-model specific keyword

        Returns:
            Prediction

        Raises:
            ValueError in case no model is trained for the requested quantile

        """
        # Check if model is trained for this quantile
        if quantile not in self.quantiles:
            raise ValueError("No model trained for requested quantile!")

        # Check/validate input
        check_array(x, force_all_finite="allow-nan")
        check_is_fitted(self)

        # Convert array to dmatrix
        dmatrix_input = xgb.DMatrix(x.copy(deep=True))

        return self.estimators_[quantile].predict(
            dmatrix_input,
            iteration_range=(0, self.estimators_[quantile].best_iteration + 1),
        )

    @classmethod
    def get_feature_importances_from_booster(cls, booster: Booster) -> np.ndarray:
        """Gets feauture importances from a XGB booster.

        This is based on the feature_importance_ property defined in:
        https://github.com/dmlc/xgboost/blob/master/python-package/xgboost/sklearn.py.

        Args:
            booster: Booster object,
                most of the times the median model (quantile=0.5) is preferred

        Returns:
            Ndarray with normalized feature importances.

        """
        # Get score
        score = booster.get_score(importance_type="gain")

        # Get feature names from booster
        feature_names = booster.feature_names

        # Get importance
        feature_importance = [score.get(f, 0.0) for f in feature_names]
        # Convert to array
        features_importance_array = np.array(feature_importance, dtype=np.float32)

        total = features_importance_array.sum()  # For normalizing
        if total == 0:
            return features_importance_array
        return features_importance_array / total  # Normalize

    @property
    def feature_names(self):
        return self._Booster.feature_names

    @property
    def can_predict_quantiles(self):
        return True



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

PROJECT_ROOT = Path(__file__).parent.parent.absolute()

try:
    __version__ = version("openstef")
except PackageNotFoundError:
    # package is not installed
    pass
